//! I2C driver for my amlib i2c interface implementation.
//!
//! See `gateware/i2c.py` for register information

use crate::{read_reg, write_reg, timer::busy_wait_us};
use core::arch::asm;
use embedded_hal::blocking::i2c::{Operation, Read, SevenBitAddress, Transactional, Write};

pub const BASE: u32 = 0xF000_1800;

const REG_W: u32 = 0x00;
const FIELD_W_SCL: u32 = 0;
const FIELD_W_OE: u32 = 1;
const FIELD_W_SDA: u32 = 2;

const REG_R: u32 = 0x04;
const FIELD_R_SDA: u32 = 0;

const WAIT_PERIOD_US: u64 = (1_000_000 / 100_000) / 2;

#[derive(Clone, Copy, Debug)]
pub enum Error {
    /// I2C bus returned a NACK
    Nack,
}

pub struct LitexI2c {
    base_addr: u32,
}

use core::fmt::Write as _;

impl LitexI2c {
    pub fn new(base_addr: u32) -> Self {
        let mut device = LitexI2c { base_addr };
        device.set_scl(true);
        device.set_sda(true);
        device.set_oe(true);

        device
    }

    fn set_reg(&mut self, offset: u32, set: bool) {
        static mut reg_state: u32 = 0;

        unsafe {
            if set {
                reg_state |= (1 << offset);
            } else {
                reg_state &= !(1 << offset);
            }

            write_reg(self.base_addr + REG_W, reg_state);
        }
    }

    fn set_scl(&mut self, set: bool) {
        defmt::debug!("set scl {}", set);
        self.set_reg(FIELD_W_SCL, set);
    }

    fn set_sda(&mut self, set: bool) {
        self.set_reg(FIELD_W_SDA, set);
    }

    fn set_oe(&mut self, set: bool) {
        self.set_reg(FIELD_W_OE, set);
    }

    fn get_sda(&mut self) -> u8 {
        unsafe { read_reg::<u8>(self.base_addr + REG_R) & (1 << FIELD_R_SDA) }
    }

    /// Generate a START condition. The next operation should be a clock
    /// transition to low.
    ///
    /// Assumes SCL and SDA have been high due to bus inactivity
    unsafe fn start(&mut self) {
        self.set_scl(true);
        self.set_sda(true);
        busy_wait_us(WAIT_PERIOD_US);

        self.set_sda(false);
        busy_wait_us(WAIT_PERIOD_US);
    }

    /// Generate a repeated START condition.
    ///
    /// Assumes it was called after an ACK bit
    unsafe fn repeated_start(&mut self) {
        self.set_scl(false);
        self.set_sda(true);
        busy_wait_us(WAIT_PERIOD_US);

        self.set_scl(true);
        busy_wait_us(WAIT_PERIOD_US);

        // TODO does this third half-clock break timing?
        self.set_sda(false);
        busy_wait_us(WAIT_PERIOD_US);
    }

    /// Handles STOP conditions.
    ///
    /// Assumes it was called after an ACK bit
    unsafe fn stop(&mut self) {
        self.set_scl(false);
        self.set_sda(false);
        busy_wait_us(WAIT_PERIOD_US);

        self.set_scl(true);
        busy_wait_us(WAIT_PERIOD_US);

        // TODO does this third half-clock break timing?
        //self.set_oe(false); // Disable and let pullup take bus high
        self.set_sda(true);
        busy_wait_us(WAIT_PERIOD_US);
    }

    unsafe fn write_byte(&mut self, byte: u8) {
        for i in 0..8 {
            // Write data out before we clock high
            self.set_scl(false);
            busy_wait_us(WAIT_PERIOD_US);

            // Clock data in
            self.set_sda((byte >> (7 - i)) & 0x01 != 0);
            self.set_scl(true);
            busy_wait_us(WAIT_PERIOD_US)
        }
    }

    unsafe fn read_byte(&mut self) -> u8 {
        let mut out = 0;
        for i in 0..8 {
            self.set_scl(false);
            busy_wait_us(WAIT_PERIOD_US);

            self.set_scl(true);
            busy_wait_us(WAIT_PERIOD_US);

            // TODO sampling at the end of the clock may not work 100% of the time
            // ideally should sample in the middle
            out |= (self.get_sda() << (7 - i));
        }

        out
    }

    // Returns if there was an ACK
    unsafe fn get_ack(&mut self) -> bool {
        self.set_scl(false);
        self.set_sda(true);
        busy_wait_us(WAIT_PERIOD_US);

        self.set_scl(true);
        busy_wait_us(WAIT_PERIOD_US);

        self.get_sda() == 0
    }

    unsafe fn set_ack(&mut self, ack: bool) {
        self.set_scl(false);
        busy_wait_us(WAIT_PERIOD_US);

        self.set_sda(!ack);
        self.set_scl(true);
        busy_wait_us(WAIT_PERIOD_US);
    }
}

impl Write<SevenBitAddress> for LitexI2c {
    // TODO errors
    type Error = Error;

    fn write(&mut self, address: SevenBitAddress, bytes: &[u8]) -> Result<(), Self::Error> {
        unsafe {
            self.start();

            // Write address and write bit
            self.write_byte((address << 1) | 0);
            if !self.get_ack() {
                defmt::debug!("write addr fail");
                self.stop();
                return Err(Error::Nack);
            }

            for b in bytes {
                self.write_byte(*b);
                defmt::debug!("write data ");
                if !self.get_ack() {
                    defmt::debug!("write data fail");
                    self.stop();
                    return Err(Error::Nack);
                }
            }

            self.stop();

            Ok(())
        }
    }
}

impl Read<SevenBitAddress> for LitexI2c {
    // TODO errors
    type Error = Error;

    fn read(&mut self, address: SevenBitAddress, buffer: &mut [u8]) -> Result<(), Self::Error> {
        unsafe {
            self.start();

            self.write_byte((address << 1) | 1);
            if !self.get_ack() {
                self.stop();
                return Err(Error::Nack);
            }

            for b in buffer {
                *b = self.read_byte();
                self.set_ack(true);
            }

            self.stop();
            Ok(())
        }
    }
}

impl Transactional for LitexI2c {
    type Error = Error;

    fn exec<'a>(
        &mut self,
        address: u8,
        operations: &mut [Operation<'a>],
    ) -> Result<(), Self::Error> {
        unsafe {
            let mut last_operation_was_read: Option<bool> = None;

            for op in operations {
                let op_is_read = match op {
                    Operation::Write(_) => false,
                    Operation::Read(_) => true,
                };

                match last_operation_was_read {
                    None => {
                        // First operation; send regular start and address
                        self.start();
                        self.write_byte((address << 1) | op_is_read as u8);
                        if !self.get_ack() {
                            self.stop();
                            return Err(Error::Nack);
                        }
                    }
                    Some(last_op_was_read) => {
                        // Not the first, send repeated start
                        self.repeated_start();

                        if last_op_was_read != op_is_read {
                            // Operations don't match, so we need to send address byte
                            self.write_byte((address << 1) | op_is_read as u8);
                            if !self.get_ack() {
                                self.stop();
                                return Err(Error::Nack);
                            }
                        }
                    }
                }

                match op {
                    Operation::Read(buffer) => {
                        for b in buffer.iter_mut() {
                            *b = self.read_byte();
                            self.set_ack(true);
                        }
                    }
                    Operation::Write(buffer) => {
                        for b in buffer.iter() {
                            self.write_byte(*b);
                            if !self.get_ack() {
                                self.stop();
                                return Err(Error::Nack);
                            }
                        }
                    }
                }

                last_operation_was_read = Some(op_is_read);
            }

            self.stop();
        }
        Ok(())
    }
}
