//! Simple module to use LiteX TIMER0 for simple timekeeping

use crate::{read_reg, write_reg};
use core::arch::asm;

const BASE: u32 = 0xf000_4000;

/// Initializes the timer to use for polling events as a wall clock.
///
/// Timer is running at system frequency (60MHz)
pub fn init() {
    unsafe {
        write_reg(BASE + 0x08, 0u32); // Disable timer
        write_reg(BASE + 0x00, 0u32); // Set LOAD value
        write_reg(BASE + 0x04, 60_000_000u32); // Set RELOAD value
        write_reg(BASE + 0x08, 1u32); // Enable timer
    }
}

// Storage for overflow time
static mut SECONDS: u32 = 0;

/// Processes potential timer events. Must be polled regularly
pub fn poll() {
    unsafe {
        if read_reg::<u32>(BASE + 0x18) == 0 {
            // No event yet, continue
            return;
        }

        // Clear TIMER0 event status, and update time
        write_reg(BASE + 0x18, 1u32);
        SECONDS += 1;
    }
}

/// Gets system time in milliseconds
pub fn millis() -> u32 {
    riscv::interrupt::free(|| {
        unsafe {
            // Latch timer value
            write_reg(BASE + 0x0C, 1u32);
            // Read timer value
            let val: u32 = read_reg(BASE + 0x10);
            let val = 60_000_000 - val;
            (SECONDS * 1000) + val / 60_000
        }
    })
}

/// Gets system time in microseconds
pub fn micros() -> u64 {
    riscv::interrupt::free(|| {
        unsafe {
            // Latch timer value
            write_reg(BASE + 0x0C, 1u32);
            // Read timer value
            let val: u32 = read_reg(BASE + 0x10);
            let val = 60_000_000 - val;
            (SECONDS as u64 * 1_000_000) + (val as u64 / 60)
        }
    })
}

pub fn busy_wait_ms(ms: u32) {
    let start = millis();
    while millis() - start < ms {
        unsafe {
            asm!("nop");
        }
    }
}

pub fn busy_wait_us(us: u64) {
    let start = micros();
    while micros() - start < us {
        unsafe {
            asm!("nop");
        }
    }
}
